/* Access.c b1.1 */

/*
	This is part of ODS2 written by Paul Nankervis,
	email address:  Paulnank@au1.ibm.com

	ODS2 is distributed freely for all members of the
	VMS community to use. However all derived works
	must maintain comments in their source to acknowledge
	the contibution of the original author.
*/

/* This module should have better checking in window
   mapping - window problems could theoretically cause
   an infinite loop :-(    */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include "ssdef.h"
#include "access.h"
#include "phy.h"


#define DEBUG_off on






/* checksum: to produce block checksum values... */

unsigned short checksum(unsigned short *block)
{
    register words = 255;
    register unsigned result = 0;
    do {
	result += *block++;
    } while (--words > 0);
    return result;
}




unsigned deaccesshead(struct VIOC *vioc)
{
#ifdef DEBUG
    printf("Deaccessing header\n");
#endif
    return cacheuntouch(&vioc->cache,1);
}

/* accesshead: find file or extension header from INDEXF... */

unsigned accesshead(struct VCB *vcb,struct fiddef *fid,unsigned extenflag,
		    struct VIOC **vioc,struct HEAD **headbuff)
{
    register unsigned sts;
    register struct VCBDEV *vcbdev;
    register unsigned idxblk = fid->fid$w_num + (fid->fid$b_nmx << 16) - 1;
    if (fid->fid$b_rvn > vcb->devices) return SS$_NOSUCHFILE;
    if (fid->fid$b_rvn < 2) {
	vcbdev = vcb->vcbdev;
    } else {
	vcbdev = &vcb->vcbdev[fid->fid$b_rvn - 1];
    }
    if (vcbdev->dev == NULL) return SS$_NOSUCHFILE;
#ifdef DEBUG
    printf("Accessing header (%d,%d,%d)\n",(fid->fid$b_nmx << 16) +
	   fid->fid$w_num,fid->fid$w_seq,fid->fid$b_rvn);
#endif
    idxblk += vcbdev->home.hm2$w_ibmapvbn + vcbdev->home.hm2$w_ibmapsize;
    if (idxblk >= swapw(vcbdev->idxfcb->head->fh2$w_recattr.fat$l_efblk)) return SS$_NOSUCHFILE;
    if ((sts = accesschunk(vcbdev->idxfcb,idxblk,vioc,(char **) headbuff,NULL)) & 1) {
	register struct HEAD *head = *headbuff;
	if (checksum((u_word *) head) != head->fh2$w_checksum) {
#ifdef DEBUG
	    printf("Accesshead checksum %d != %d\n",checksum((u_word *) head),head->fh2$w_checksum);
#endif
	    sts = SS$_NOSUCHFILE;
	} else {
	    if (head->fh2$w_fid.fid$w_num != fid->fid$w_num ||
		head->fh2$w_fid.fid$b_nmx != fid->fid$b_nmx ||
		head->fh2$w_fid.fid$w_seq != fid->fid$w_seq ||
		(head->fh2$w_fid.fid$b_rvn != fid->fid$b_rvn &&
		 head->fh2$w_fid.fid$b_rvn != 0)) {
#ifdef DEBUG
		printf("Accesshead fileid doesn't match %d %d %d\n",
		       fid->fid$w_num,fid->fid$w_seq,fid->fid$b_rvn);
#endif
		sts = SS$_NOSUCHFILE;
	    } else {
		if (head->fh2$b_idoffset < 38 ||
		    head->fh2$b_idoffset > head->fh2$b_mpoffset ||
		    head->fh2$b_mpoffset > head->fh2$b_acoffset ||
		    head->fh2$b_acoffset > head->fh2$b_rsoffset ||
		    head->fh2$b_map_inuse > (head->fh2$b_acoffset - head->fh2$b_mpoffset)) {
#ifdef DEBUG
		    printf("Accesshead areas incorrect\n");
#endif
		    sts = SS$_NOSUCHFILE;
		}
	    }
	}
    }
    return sts;
}


void deltree(struct CACHE *cacheobj)
{
    if (cacheobj != NULL) {
	if (cacheobj->left != NULL) deltree(cacheobj->left);
	if (cacheobj->right != NULL) deltree(cacheobj->right);
	cachefree(cacheobj,0);
    }
}

/* checkcount: determine tree reference count... */

unsigned checkcount(struct CACHE *cacheobj)
{
    register refcount = 0;
    if (cacheobj != NULL) {
	if (cacheobj->left != NULL) refcount += checkcount(cacheobj->left);
	refcount += cacheobj->refcount;
	if (cacheobj->right != NULL) refcount += checkcount(cacheobj->right);
    }
    return refcount;
}


/* deaccessfile: finish accessing a file.... */

unsigned deaccessfile(struct FCB *fcb)
{
#ifdef DEBUG
    printf("Deaccessing file (%x) reference %d\n",fcb->fileno,fcb->cache.refcount);
#endif
    if (fcb->cache.refcount == 1) {
	register unsigned refcount;
	refcount = checkcount((struct CACHE *) fcb->wcb) +
	    checkcount((struct CACHE *) fcb->vioc);
	if (refcount != 0) {
#ifdef DEBUG
	    printf("File reference counts non-zero %d %d\n",
		   checkcount((struct CACHE *) fcb->wcb),checkcount((struct CACHE *) fcb->vioc));
#endif
	    return SS$_BUGCHECK;
	}
    }
    cacheuntouch(&fcb->cache,0);
    return SS$_NORMAL;
}


struct CACHE *accessfileproxy(struct CACHE *cacheobj)
{
    register struct FCB *fcb = (struct FCB *) cacheobj;
    if (fcb->vioc != NULL) return &fcb->vioc->cache;
    if (fcb->wcb != NULL) return &fcb->wcb->cache;
    return cacheobj;
}


int filecmp(unsigned keylen,void *key,void *node)
{
    register struct FCB *fcb = (struct FCB *) node;
    if (keylen == fcb->fileno) {
	return 0;
    } else {
	if (keylen < fcb->fileno) {
	    return -1;
	} else {
	    return 1;
	}
    }
}



/* accessfile: open up file for access... */

unsigned accessfile(struct VCB *vcb,struct fiddef *fid,struct FCB **fcbadd)
{
    register struct FCB *fcb;
    unsigned create = sizeof(struct FCB);
    register unsigned fileno = (fid->fid$b_nmx << 16) + fid->fid$w_num;
#ifdef DEBUG
    printf("Accessing file (%d,%d,%d)\n",(fid->fid$b_nmx << 16) +
	   fid->fid$w_num,fid->fid$w_seq,fid->fid$b_rvn);
#endif
    if (fileno < 1) return SS$_BADPARAM;
    if (fid->fid$b_rvn > 1) fileno += ((fid->fid$b_rvn - 1) << 24);
    fcb = cachesearch((void *) &vcb->fcb,fileno,NULL,filecmp,&create);
    if (fcb == NULL) return SS$_INSFMEM;
    /* If not found make one... */
    if (create == 0) {
	register unsigned sts;
	fcb->fileno = fileno;
	fcb->status = 0;
	fcb->rvn = fid->fid$b_rvn;
	if (fcb->rvn == 0 && vcb->devices > 1) fcb->rvn = 1;
	fcb->vcb = vcb;
	fcb->wcb = NULL;
	fcb->vioc = NULL;
	fcb->cache.getproxy = accessfileproxy;
	if (vcb->idxboot != NULL) {
	    *fcbadd = fcb;
	    fcb->hiblock = 32767;	/* guess at indexf.sys file size */
	    fcb->highwater = 0;
	    fcb->head = vcb->idxboot;	/* Load bootup header */
	}
	sts = accesshead(vcb,fid,0,&fcb->headvioc,&fcb->head);
	if (sts & 1) {
	    fcb->hiblock = swapw(fcb->head->fh2$w_recattr.fat$l_hiblk);
	    if (fcb->head->fh2$b_idoffset > 39) {
		fcb->highwater = fcb->head->fh2$l_highwater;
	    } else {
		fcb->highwater = 0;
	    }
	} else {
	    cachefree((struct CACHE *) fcb,1);
	}
    }
    *fcbadd = fcb;
    return SS$_NORMAL;
}


/* dismount: finish processing on a volume */

unsigned dismount(struct VCB *vcb)
{
    register unsigned sts,device;
    struct VCBDEV *vcbdev;
#ifdef DEBUG
    printf("Dismounting disk %d\n",checkcount((struct CACHE *) vcb->fcb));
#endif
    sts = SS$_NORMAL;
    if (checkcount((struct CACHE *) vcb->fcb) != vcb->devices) {
	sts = SS$_ABORT;
    } else {
	vcbdev = vcb->vcbdev;
	for (device = 0; device < vcb->devices; device++) {
	    if (vcbdev->dev != NULL) {
		if (checkcount((struct CACHE *) vcbdev->idxfcb->wcb) +
		    checkcount((struct CACHE *) vcbdev->idxfcb->vioc) != 1) {
		    sts = SS$_BUGCHECK;
		} else {
		    deaccesshead(vcbdev->idxfcb->headvioc);
		    cacheuntouch(&vcbdev->idxfcb->cache,0);
		    vcbdev->dev->vcb = NULL;
		}
	    }
	    vcbdev++;
	}
	deltree((struct CACHE *) vcb->dircache);
	deltree((struct CACHE *) vcb->fcb);
	free(vcb);
    }
    return sts;
}

/* mount: make disk volume available for processing... */

unsigned mount(unsigned flags,unsigned devices,char *devnam[],char *label[],struct VCB **retvcb)
{
    register unsigned device,sts;
    struct VCB *vcb;
    struct VCBDEV *vcbdev;
    if (sizeof(struct HOME) != 512 || sizeof(struct HEAD) != 512) return SS$_NOTIMPLEM;
    vcb = (struct VCB *) malloc(sizeof(struct VCB) + (devices - 1) * sizeof(struct VCBDEV));
    if (vcb == NULL) return SS$_INSFMEM;
    vcb->status = 0;
    vcb->fcb = NULL;
    vcb->dircache = NULL;
    vcbdev = vcb->vcbdev;
    for (device = 0; device < devices; device++) {
	sts = SS$_NOSUCHVOL;
	vcbdev->dev = NULL;
	if (strlen(devnam[device])) {
	    sts = device_lookup(strlen(devnam[device]),devnam[device],1,&vcbdev->dev);
	    if (sts & 1) {
		int hba = 1;
		do {
		    if ((sts = phyio_read(vcbdev->dev->handle,hba,sizeof(struct HOME),(char *) &vcbdev->home)) & 1) {
			if (memcmp(vcbdev->home.hm2$t_format,"DECFILE11B  ",12) != 0) sts = SS$_BADPARAM;
		    }
		} while ((sts & 1) == 0 && (hba += 1) < 100);
		if (sts & 1) {
		    if (vcbdev->home.hm2$w_checksum2 != checksum((unsigned short *) &vcbdev->home)) {
			sts = SS$_DATACHECK;
		    } else {
			struct HEAD idxboot;	/* Local for bootstrapping volume */
			struct fiddef idxfid = {1,1,0,0};
			idxfid.fid$b_rvn = device + 1;
			if ((sts = phyio_read(vcbdev->dev->handle,vcbdev->home.hm2$l_ibmaplbn +
					      vcbdev->home.hm2$w_ibmapsize,sizeof(struct HEAD),(char *) &idxboot)) & 1) {
			    if (idxboot.fh2$w_fid.fid$w_num != idxfid.fid$w_num ||
				idxboot.fh2$w_fid.fid$b_nmx != idxfid.fid$b_nmx ||
				idxboot.fh2$w_fid.fid$w_seq != idxfid.fid$w_seq ||
				idxboot.fh2$w_fid.fid$b_rvn != 0 && idxboot.fh2$w_fid.fid$b_rvn != idxfid.fid$w_seq)
				sts = SS$_DATACHECK;
			    if (idxboot.fh2$w_checksum != checksum((unsigned short *) &idxboot)) sts = SS$_DATACHECK;
			}
			vcbdev->idxfcb = NULL;
			if (sts & 1) {
			    vcb->devices = device + 1;
			    vcb->idxboot = &idxboot;	/* For getwindow to find index file*/
			    sts = accessfile(vcb,&idxfid,&vcbdev->idxfcb);
			    if (sts & 1) vcbdev->dev->vcb = vcb;
			}
		    }
		}
	    }
	    if ((sts & 1) == 0) vcbdev->dev = NULL;
	}
	if (device == 0 && vcbdev->dev == NULL) {
	    free(vcb);
	    return sts;
	}
	vcbdev++;
    }
    vcb->idxboot = NULL;
    if (retvcb != NULL) *retvcb = vcb;
    return sts;
}


int wincmp(unsigned keylen,void *key,void *node)
{
    register struct WCB *wcb = (struct WCB *) node;
    if (keylen < wcb->loblk) {
	return -1;
    } else {
	if (keylen <= wcb->hiblk) {
	    return 0;
	} else {
	    register struct WCB **prev_wcb = (struct WCB **) key;
	    if (*prev_wcb == NULL) {
		*prev_wcb = wcb;
	    } else {
		if ((*prev_wcb)->hiblk < wcb->hiblk) *prev_wcb = wcb;
	    }
	    return 1;
	}
    }
}


/* getwindow: find a window to map VBN to LBN ... */

unsigned getwindow(struct FCB *fcb,unsigned vbn,unsigned *phyrvn,unsigned *phyblk,unsigned *phylen)
{
    register struct WCB *wcb;
    struct WCB *prev_wcb = NULL;
    unsigned create = sizeof(struct WCB);
#ifdef DEBUG
    printf("Accessing window for vbn %d, file (%x)\n",vbn,fcb->fileno);
#endif
    wcb = cachesearch((void *) &fcb->wcb,vbn,&prev_wcb,wincmp,&create);
    if (wcb == NULL) return SS$_INSFMEM;
    /* If not found make one... */
    if (create == 0) {
	register unsigned wd_base,wd_exts;
	unsigned prev_hiblk,rvn;
	struct VIOC *vioc;
	struct HEAD *head;
	vioc = NULL;
	wd_base = 1;
	rvn = fcb->rvn;
	head = fcb->head;
	prev_hiblk = 0;
	if (prev_wcb != NULL) {
	    register unsigned sts;
	    register struct fiddef *fid = &prev_wcb->hd_fid;
	    register struct fiddef *filefid = &fcb->head->fh2$w_fid;
	    if (fid->fid$w_num != filefid->fid$w_num ||
		fid->fid$b_nmx != filefid->fid$b_nmx ||
		((fid->fid$b_rvn > 1 || fcb->rvn > 1) && fid->fid$b_rvn != fcb->rvn)) {
		wd_base = prev_wcb->hd_base;
		rvn = prev_wcb->hd_fid.fid$b_rvn;
		sts = accesshead(fcb->vcb,fid,1,&vioc,&head);
		if ((sts & 1) == 0) {
		    cachefree((struct CACHE *) wcb,0);
		    return sts;
		}
	    }
	    prev_hiblk = prev_wcb->hiblk;
	}
	wcb->hd_base = wd_base;
	wd_exts = 0;
#ifdef DEBUG
	printf("Making window %d %d\n",wd_base,prev_hiblk);
#endif
	do {
	    register unsigned short *mp = (unsigned short *) head + head->fh2$b_mpoffset;
	    register unsigned short *me = mp + head->fh2$b_map_inuse;
	    while (mp < me) {
		register unsigned phylen,phyblk;
		switch ((*mp) >> 14) {
		    case 0:
			phylen = 0;
			mp++;
			break;
		    case 1:
			phylen = ((*mp) & 0377) + 1;
			phyblk = (((*mp) & 037400) << 8) + mp[1];
			mp += 2;
			break;
		    case 2:
			phylen = ((*mp) & 037777) + 1;
			phyblk = (mp[2] << 16) + mp[1];
			mp += 3;
			break;
		    case 3:
			phylen = (((*mp) & 037777) << 16) + mp[1] + 1;
			phyblk = (mp[3] << 16) + mp[2];
			mp += 4;
		}
		if (phylen > 0 && wd_base > prev_hiblk) {
		    register struct EXT *ext;
		    if (wd_exts == 0) wcb->loblk = wd_base;
		    ext = &wcb->ext[wd_exts++];
		    ext->phylen = phylen;
		    ext->phyblk = phyblk;
		    wd_base += phylen;
		    if (wd_exts >= EXTMAX) {
			if (wd_base > vbn) {
			    break;
			} else {
			    wd_exts = 0;
			}
		    }
		} else {
		    wd_base += phylen;
		}
	    }
	    if (wd_base > vbn) {
		break;
	    } else {
		register unsigned sts;
		struct fiddef extfid;
		memcpy(&extfid,&head->fh2$w_ext_fid,sizeof(struct fiddef));
		if (extfid.fid$b_rvn != 0 && extfid.fid$b_rvn != rvn) {
		    wd_exts = 0;/* Can't let window extend across devices */
		    rvn = extfid.fid$b_rvn;
		} else {
		    extfid.fid$b_rvn = rvn;
		}
		if (vioc != NULL) deaccesshead(vioc);
		sts = accesshead(fcb->vcb,&extfid,1,&vioc,&head);
		if ((sts & 1) == 0) {
		    cachefree((struct CACHE *) wcb,0);
		    return sts;
		}
		wcb->hd_base = wd_base;
	    }
	} while (wd_base <= vbn);
	memcpy(&wcb->hd_fid,&head->fh2$w_fid,sizeof(struct fiddef));
	wcb->hd_fid.fid$b_rvn = rvn;
	wcb->hiblk = wd_base - 1;
	wcb->extcount = wd_exts;
	if (vioc != NULL) deaccesshead(vioc);
    } {
	register struct EXT *ext = wcb->ext;
	register unsigned extcnt = wcb->extcount;
	register unsigned togo = vbn - wcb->loblk;
	while (togo >= ext->phylen) {
	    togo -= (ext++)->phylen;
	    if (extcnt-- < 1) return RMS$_BUG;
	}
	*phyrvn = wcb->hd_fid.fid$b_rvn;
	*phyblk = ext->phyblk + togo;
	*phylen = ext->phylen - togo;
#ifdef DEBUG
	printf("Mapping vbn %d to %d (%d -> %d)[%d] file (%x)\n",
	       vbn,*phyblk,wcb->loblk,wcb->hiblk,wcb->hd_base,fcb->fileno);
#endif
	cacheuntouch(&wcb->cache,1);
    }
    return SS$_NORMAL;
}


unsigned deaccesschunk(struct VIOC *vioc,int reuse)
{
    return cacheuntouch(&vioc->cache,reuse);
}


int chunkcmp(unsigned keylen,void *key,void *node)
{
    register struct VIOC *vioc = (struct VIOC *) node;
    if (keylen < vioc->base) {
	return -1;
    } else {
	if (keylen >= vioc->base + CHUNKSIZE) {
	    return 1;
	} else {
	    return 0;
	}
    }
}


/* accesschunk: find memory pointer to a 'chunk' of a file ... */

unsigned accesschunk(struct FCB *fcb,unsigned vbn,struct VIOC **retvioc,char **retbuff,unsigned *retblocks)
{
    /* First find cache entry... */
    register struct VIOC *vioc;
    unsigned create = sizeof(struct VIOC);
#ifdef DEBUG
    printf("Access chunk %d (%x)\n",vbn,fcb->fileno);
#endif
    if (vbn < 1 || vbn > fcb->hiblock) return SS$_ENDOFFILE;
    vioc = cachesearch((void *) &fcb->vioc,vbn,NULL,chunkcmp,&create);
    if (vioc == NULL) return SS$_INSFMEM;
    /* If not found make one... */
    if (create == 0) {
	register unsigned base,length;
	register char *address;
	vioc->base = base = (vbn - 1) / CHUNKSIZE * CHUNKSIZE + 1;
	length = fcb->hiblock - base + 1;
	if (length > CHUNKSIZE) length = CHUNKSIZE;
	address = (char *) vioc->data;
	do {
	    if (fcb->highwater > 0 && base >= fcb->highwater) {
		memset(address,0,length * 512);
		length = 0;
	    } else {
		register unsigned sts;
		unsigned rvn,mapblk,maplen;
		register struct VCBDEV *vcbdev;
		sts = getwindow(fcb,base,&rvn,&mapblk,&maplen);
		if (sts & 1) {
		    if (maplen > length) maplen = length;
		    if (fcb->highwater > 0 && base + maplen > fcb->highwater) {
			maplen = fcb->head->fh2$l_highwater - base;
		    }
		    if (rvn > fcb->vcb->devices) {
			sts = SS$_NOSUCHFILE;
		    } else {
			if (rvn < 2) {
			    vcbdev = fcb->vcb->vcbdev;
			} else {
			    vcbdev = &fcb->vcb->vcbdev[rvn - 1];
			}
			if (vcbdev->dev == NULL) return SS$_NOSUCHFILE;
			sts = phyio_read(vcbdev->dev->handle,mapblk,maplen * 512,address);
		    }
		}
		if ((sts & 1) == 0) {
		    cachefree((struct CACHE *) vioc,1);
		    return sts;
		}
		length -= maplen;
		base += maplen;
		address += maplen * 512;
	    }
	} while (length > 0);
    }
    /* Return result to caller.. */
    *retvioc = vioc;
    *retbuff = vioc->data[vbn - vioc->base];
    if (retblocks != NULL) {
	unsigned blocks = vioc->base + CHUNKSIZE - vbn;
	if (blocks > fcb->hiblock - vbn) {
	    *retblocks = fcb->hiblock - vbn + 1;
	} else {
	    *retblocks = blocks;
	}
    }
    return SS$_NORMAL;
}
