/* Access.h b1.1    Definitions for file access routines */

/*
	This is part of ODS2 written by Paul Nankervis,
	email address:  Paulnank@au1.ibm.com

	ODS2 is distributed freely for all members of the
	VMS community to use. However all derived works
	must maintain comments in their source to acknowledge
	the contibution of the original author.
*/


#include "cache.h"

#define swapw(w) (w[0]<<16 | w[1])

typedef unsigned char u_byte;
typedef unsigned short u_word;
typedef unsigned int u_long;

struct UIC {
    u_word u_grp;
    u_word u_mem;
};

struct TIME {
    u_byte time[8];
};				/* structure of time - albeit the simple version! */

struct fiddef {
    u_word fid$w_num;
    u_word fid$w_seq;
    u_byte fid$b_rvn;
    u_byte fid$b_nmx;
};


struct RECATTR {
    u_byte fat$b_rtype;
    u_byte fat$b_rattrib;
    u_word fat$w_rsize;
    u_word fat$l_hiblk[2];
    u_word fat$l_efblk[2];
    u_word fat$w_ffbyte;
    u_byte fat$b_bktsize;
    u_byte fat$b_vfcsize;
    u_word fat$w_maxrec;
    u_word fat$w_defext;
    u_word fat$w_gbc;
    u_byte fat$_UU0[8];
    u_word fat$w_versions;
};

struct HOME {
    u_long hm2$l_homelbn;
    u_long hm2$l_alhomelbn;
    u_long hm2$l_altidxlbn;
    u_word hm2$w_struclev;
    u_word hm2$w_cluster;
    u_word hm2$w_homevbn;
    u_word hm2$w_alhomevbn;
    u_word hm2$w_altidxvbn;
    u_word hm2$w_ibmapvbn;
    u_long hm2$l_ibmaplbn;
    u_long hm2$l_maxfiles;
    u_word hm2$w_ibmapsize;
    u_word hm2$w_resfiles;
    u_word hm2$w_devtype;
    u_word hm2$w_rvn;
    u_word hm2$w_setcount;
    u_word hm2$w_volchar;
    struct UIC hm2$w_volowner;
    u_long hm2$l_reserved1;
    u_word hm2$w_protect;
    u_word hm2$w_fileprot;
    u_word hm2$w_reserved2;
    u_word hm2$w_checksum1;
    struct TIME hm2$q_credate;
    u_byte hm2$b_window;
    u_byte hm2$b_lru_lim;
    u_word hm2$w_extend;
    struct TIME hm2$q_retainmin;
    struct TIME hm2$q_retainmax;
    struct TIME hm2$q_revdate;
    u_byte hm2$r_min_class[20];
    u_byte hm2$r_max_class[20];
    u_byte hm2$t_reserved3[320];
    u_long hm2$l_serialnum;
    char hm2$t_strucname[12];
    char hm2$t_volname[12];
    char hm2$t_ownername[12];
    char hm2$t_format[12];
    u_word hm2$w_reserved4;
    u_word hm2$w_checksum2;
};

struct IDENT {
    char fi2$t_filename[20];
    u_word fi2$w_revision;
    struct TIME fi2$q_credate;
    struct TIME fi2$q_revdate;
    struct TIME fi2$q_expdate;
    struct TIME fi2$q_bakdate;
    char fi2$t_filenamext[66];
};

struct HEAD {
    u_byte fh2$b_idoffset;
    u_byte fh2$b_mpoffset;
    u_byte fh2$b_acoffset;
    u_byte fh2$b_rsoffset;
    u_word fh2$w_seg_num;
    u_word fh2$w_struclev;
    struct fiddef fh2$w_fid;
    struct fiddef fh2$w_ext_fid;
    struct RECATTR fh2$w_recattr;
    u_long fh2$l_filechar;
    u_word fh2$w_reserved1;
    u_byte fh2$b_map_inuse;
    u_byte fh2$b_acc_mode;
    struct UIC fh2$l_fileowner;
    u_word fh2$w_fileprot;
    struct fiddef fh2$w_backlink;
    u_byte fh2$b_journal;
    u_byte fh2$b_ru_active;
    u_word fh2$w_reserved2;
    u_long fh2$l_highwater;
    u_byte fh2$b_reserved3[8];
    u_byte fh2$r_class_prot[20];
    u_byte fh2$r_restofit[402];
    u_word fh2$w_checksum;
};

struct dir$rec {
    u_word dir$size;
    u_word dir$verlimit;
    u_byte dir$flags;
    u_byte dir$namecount;
    char dir$name[1];
};

struct dir$ent {
    u_word dir$version;
    struct fiddef dir$fid;
};

struct EXT {
    unsigned phylen;
    unsigned phyblk;
};

#define EXTMAX 20

struct WCB {
    struct CACHE cache;
    unsigned loblk,hiblk;
    unsigned hd_base;
    unsigned extcount;
    struct EXT ext[EXTMAX];
    struct fiddef hd_fid;
};


#define CHUNKSIZE 4

struct VIOC {
    struct CACHE cache;
    unsigned status;
    unsigned modmap;
    unsigned base;
    char data[CHUNKSIZE][512];
};

struct FCB {
    struct CACHE cache;
    struct VCB *vcb;
    struct VIOC *headvioc;
    struct HEAD *head;
    struct WCB *wcb;
    struct VIOC *vioc;
    unsigned fileno;
    unsigned hiblock;
    unsigned highwater;
    unsigned short status;
    unsigned char rvn;
};

struct DIRCACHE {
    struct CACHE cache;
    int dirlen;
    struct fiddef dirid;
    char dirnam[1];
};

struct VCB {
    unsigned status;
    unsigned devices;
    struct HEAD *idxboot;
    struct FCB *fcb;
    struct DIRCACHE *dircache;
    struct VCBDEV {
	struct DEV *dev;
	struct FCB *idxfcb;
	struct HOME home;
    } vcbdev[1];
};

struct DEV {
    struct CACHE cache;
    unsigned handle;
    struct VCB *vcb;
    int devlen;
    char devnam[1];
};

unsigned device_lookup(unsigned devlen,char *devnam,int create,struct DEV **retdev);

unsigned dismount(struct VCB *vcb);
unsigned mount(unsigned flags,unsigned devices,char *devnam[],char *label[],struct VCB **vcb);

unsigned deaccessfile(struct FCB *fcb);
unsigned accessfile(struct VCB *vcb,struct fiddef *fid,struct FCB **fcb);

unsigned deaccesschunk(struct VIOC *vioc,int reuse);
unsigned accesschunk(struct FCB *fcb,unsigned vbn,struct VIOC **retvioc,char **retbuff,unsigned *retblocks);
