/* Direct.c b1.1 */

/*
	This is part of ODS2 written by Paul Nankervis,
	email address:  Paulnank@au1.ibm.com

	ODS2 is distributed freely for all members of the
	VMS community to use. However all derived works
	must maintain comments in their source to acknowledge
	the contibution of the original author.
*/

/*  This version takes relative version from last name -
    no good if multiple name entries for a file!  */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <memory.h>
#include "ssdef.h"
#include "descrip.h"
#include "fibdef.h"
#include "access.h"

#define DEBUG_off on

#define FCH$M_DIRECTORY 0x2000

unsigned namecheck(char *str,int len,int *retlen,int *retver,int *wildflag)
{
    register char ch,*ptr = str;
    register togo = len;
    register dots = 0;
    int wildcard = 0;
    while (togo-- > 0) {
	ch = *ptr++ & 127;
	if (ch == '.') {
	    if (dots++ > 1) break;
	} else {
	    if (ch == ';') {
		break;
	    } else {
		if (ch == '*' || ch == '%') {
		    wildcard = 1;
		} else {
		    if (ch == '[' || ch == ']' || ch == ':' ||
			!isprint(ch)) return SS$_BADFILENAME;
		}
	    }
	}
    }
    if (dots == 0) return SS$_BADFILENAME;
    *retlen = len - togo - 1;
    dots = 0;
    if (togo > 0) {
	ch = *ptr;
	if (ch == '*') {
	    if (togo > 1) return SS$_BADFILENAME;
	    dots = 32768;
	    wildcard = 1;
	} else {
	    int sign = 1;
	    if (ch == '-') {
		togo--;
		ptr++;
		sign = -1;
	    }
	    while (togo-- > 0) {
		ch = *ptr++;
		if (!isdigit(ch)) return SS$_BADFILENAME;
		dots = dots * 10 + (ch - '0');
	    }
	    dots *= sign;
	}
    }
    *retver = dots;
    *wildflag = wildcard;
#ifdef DEBUG
    printf("Namecheck %d %d %d\n",*retlen,*retver,*wildflag);
#endif
    return SS$_NORMAL;
}



#define MAT_LT -1
#define MAT_EQ 0
#define MAT_GT 1
#define MAT_NE 2

int namematch(char *spec,int speclen,char *entry,int entrylen)
{
    register spclen = speclen,entlen = entrylen;
    register char *spc = spec,*ent = entry;
    /* See how much matches without wildcards... */
    while (spclen > 0 && entlen > 0) {
	    if (*spc != *ent) if (toupper(*spc) != toupper(*ent)) break;
	spclen--;
	spc++;
	entlen--;
	ent++;
    }
    /* Mismatch - return result unless wildcard... */
    if (spclen <= 0) {
	if (entlen <= 0) {
	    return MAT_EQ;
	} else {
	    return MAT_GT;
	}
    } else {
	if (*spc != '*' && *spc != '%') {
		if (entlen > 0) if (toupper(*spc) < toupper(*ent)) return MAT_GT;
	    return MAT_LT;
	}
    }
    /* See if we can find a match with wildcards */
    {
	register wildflag = 0;
	register sublen = 0;
	while (sublen < spclen || sublen < entlen) {
	    register char sch = 'a';
	    if (sublen < spclen) sch = toupper(spc[sublen]);
	    if (sch == '*') {
		spclen -= (sublen + 1);
		if (spclen <= 0) return MAT_EQ;
		spc += (sublen + 1);
		ent += sublen;
		entlen -= sublen;
		sublen = 0;
		wildflag = 1;
	    } else {
		register char ech = 'a';
		if (sublen < entlen) ech = toupper(ent[sublen]);
		if (sch == '%' || sch == ech) {
		    sublen++;
		} else {
		    if (wildflag && sublen < entlen--) {
			sublen = 0;
			ent++;
		    } else {
			return MAT_NE;
		    }
		}
	    }
	}
    }
    return MAT_EQ;
}



unsigned findfile(struct FCB *fcb,
		  struct dsc$descriptor *fibdsc,struct dsc$descriptor *filedsc,
		  unsigned short *reslen,struct dsc$descriptor *resdsc,unsigned endblk)
{
    register unsigned sts,curblk = 0;
    char *searchspec;
    int searchlen,version,wildcard,wcc_flag;
    struct fibdef *fib = (struct fibdef *) fibdsc->dsc$a_pointer;

    /* 1) Generate start block (wcc gives start point)
       2) Search for start
       3) Scan until found or too big or end   */

    if (fib->fib$l_wcc != 0) {
	searchspec = resdsc->dsc$a_pointer;
	sts = namecheck(searchspec,*reslen,&searchlen,&version,&wildcard);
	curblk = fib->fib$l_wcc;
	wcc_flag = 1;
    } else {
	searchspec = filedsc->dsc$a_pointer;
	sts = namecheck(searchspec,filedsc->dsc$w_length,&searchlen,&version,&wildcard);
	wcc_flag = 0;
    }


    /* first find start starting block...*/

    if (*searchspec == '*' || *searchspec == '%') {
	curblk = 1;
    } else {
	register unsigned loblk = 1,hiblk = endblk;
	if (curblk < 1 || curblk > endblk) curblk = (endblk + 1) / 2;
	while ((sts & 1) && loblk < hiblk) {
	    char *buffer;
	    struct VIOC *vioc;
	    register cmp;
	    register struct dir$rec *dr;
	    if ((sts = accesschunk(fcb,curblk,&vioc,&buffer,NULL)) & 1) {
		dr = (struct dir$rec *) buffer;
		cmp = namematch(searchspec,searchlen,dr->dir$name,dr->dir$namecount);
#ifdef DEBUF
		printf("Direct %6.6s %d %6.6s %d -> %d\n",
		       searchspec,searchlen,dr->dir$name,dr->dir$namecount,cmp);
#endif
		if (cmp == MAT_EQ) {
		    /*match */
		    if (version < 1 || version > 32767) {
			cmp = MAT_NE;	/* no match - want to find start */
		    } else {
			register struct dir$ent *de = (struct dir$ent *) (dr->dir$name +
									  ((dr->dir$namecount + 1) & ~1));
			if (de->dir$version < version) {
			    cmp = MAT_GT;	/* too far... */
			} else {
			    if (de->dir$version > version) {
				cmp = MAT_LT;	/* further ahead... */
			    }
			}
		    }
		}
		sts = deaccesschunk(vioc,1);
		switch (cmp) {
		    case MAT_LT:
			loblk = curblk;
			if (curblk == fib->fib$l_wcc && curblk < hiblk) {
			    curblk++;
			} else {
			    curblk = (loblk + hiblk + 1) / 2;
			}
			break;
		    case MAT_GT:
		    case MAT_NE:
			hiblk = curblk - 1;
			curblk = (loblk + hiblk + 1) / 2;
			break;
		    case MAT_EQ:
			hiblk = loblk = curblk;
		}
	    }
	}
    }

    /* Now to read sequentially to start point, then find entry... */

    while ((sts & 1) && curblk <= endblk) {
	char *buffer;
	struct VIOC *vioc;
	if ((sts = accesschunk(fcb,curblk,&vioc,&buffer,NULL)) & 1) {
	    register struct dir$rec *dr = (struct dir$rec *) buffer;
	    while ((char *) dr < buffer + 506) {
		register cmp;
		register char *nr = (char *) dr + dr->dir$size + 2;
		if (nr > buffer + 512) nr = buffer + 512;
		cmp = namematch(searchspec,searchlen,dr->dir$name,dr->dir$namecount);
#ifdef DEBUF
		printf("Direct %6.6s %d %6.6s %d -> %d\n",
		       searchspec,searchlen,dr->dir$name,dr->dir$namecount,cmp);
#endif
		if (cmp == MAT_EQ) {
		    register relver = 0;
		    register struct dir$ent *de = (struct dir$ent *) (dr->dir$name +
								      ((dr->dir$namecount + 1) & ~1));
		    cmp = MAT_LT;	/* default is no match */
		    while ((char *) de < nr) {
			if (version <= 0 || version >= de->dir$version) {
			    if (wcc_flag) {
				/* if wcc then need to switch to spec... */
				searchspec = filedsc->dsc$a_pointer;
				sts = namecheck(searchspec,filedsc->dsc$w_length,&searchlen,&version,&wildcard);
				if ((sts & 1) == 0) break;
				wcc_flag = 0;
				cmp = namematch(searchspec,searchlen,dr->dir$name,
						dr->dir$namecount);
				if (cmp != MAT_EQ) break;
			    } else {
				if (version == 32768 || version == relver || version == de->dir$version) {
				    register scale = 10;
				    register version = de->dir$version;
				    register length = dr->dir$namecount;
				    register char *ptr = resdsc->dsc$a_pointer;
				    memcpy(ptr,dr->dir$name,length);
				    while (version >= scale) scale *= 10;
				    ptr += length++;
				    *ptr++ = ';';
				    do {
					scale /= 10;
					*ptr++ = version / scale + '0';
					version %= scale;
					length++;
				    } while (scale > 1);
				    *reslen = length;
				    memcpy(&fib->fib$w_fid_num,&de->dir$fid,sizeof(struct fiddef));
				    if (fib->fib$b_fid_rvn == 0) fib->fib$b_fid_rvn = fcb->rvn;
				    if (wildcard || (fib->fib$w_nmctl & FIB$M_WILD)) {
					fib->fib$l_wcc = curblk;
				    } else {
					fib->fib$l_wcc = 0;
				    }
				    return deaccesschunk(vioc,1);
				} else {
				    if (version >= 0) {
					if (wildcard) {
					    cmp = MAT_NE;
					} else {
					    cmp = MAT_GT;
					}
					break;
				    }
				}
			    }
			}
			de++;
			relver--;
		    }
		    if ((sts & 1) == 0) break;
		}
		if (cmp == MAT_GT) {
		    if (wcc_flag) {
			/* if wcc then need to switch to spec... */
			register unsigned sts;
			searchspec = filedsc->dsc$a_pointer;
			sts = namecheck(searchspec,filedsc->dsc$w_length,&searchlen,&version,&wildcard);
			wcc_flag = 0;
		    } else {
			curblk = endblk;	/* give up */
			break;
		    }
		}
		dr = (struct dir$rec *) nr;
	    }
	    {
		register unsigned dests = deaccesschunk(vioc,1);
		if (sts & 1) sts = dests;
	    }
	    curblk++;
	}
    }
    if (sts & 1) {
	fib->fib$l_wcc = 0;
	if (wildcard) {
	    sts = SS$_NOMOREFILES;
	} else {
	    sts = SS$_NOSUCHFILE;
	}
    }
    return sts;
}


unsigned lookup(struct VCB *vcb,struct dsc$descriptor * fibdsc,
		struct dsc$descriptor *filedsc,unsigned short *reslen,
		struct dsc$descriptor *resdsc)
{
    register unsigned sts,eofblk;
    struct FCB *fcb;
    struct fibdef *fib = (struct fibdef *) fibdsc->dsc$a_pointer;
    sts = accessfile(vcb,(struct fiddef *) & fib->fib$w_did_num,&fcb);
    if (sts & 1) {
	if (fcb->head->fh2$l_filechar & FCH$M_DIRECTORY) {
	    eofblk = swapw(fcb->head->fh2$w_recattr.fat$l_efblk);
	    if (fcb->head->fh2$w_recattr.fat$w_ffbyte == 0) --eofblk;
	    sts = findfile(fcb,fibdsc,filedsc,reslen,resdsc,eofblk);
	    {
		register unsigned dests = deaccessfile(fcb);
		if (sts & 1) sts = dests;
	    }
	} else {
	    sts = SS$_BADIRECTORY;
	}
    }
    return sts;
}
