
/* PHYW95.C  b1.1    Physical I/O module for Windows 95 */

#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include <winioctl.h>
#include <string.h>
#include <ctype.h>
#include <memory.h>
#include "phy.h"

BOOL
GetDiskGeometry(
    HANDLE hDisk,
    PDISK_GEOMETRY lpGeometry
    )

{
    DWORD ReturnedByteCount;

    return DeviceIoControl(
                hDisk,
                IOCTL_DISK_GET_DRIVE_GEOMETRY,
                NULL,
                0,
                lpGeometry,
                sizeof(*lpGeometry),
                &ReturnedByteCount,
                NULL
                );
}


BOOL
LockVolume(
    HANDLE hDisk
    )
{
    DWORD ReturnedByteCount;

    return DeviceIoControl(
                hDisk,
                FSCTL_LOCK_VOLUME,
                NULL,
                0,
                NULL,
                0,
                &ReturnedByteCount,
                NULL
                );
}

BOOL
UnlockVolume(
    HANDLE hDisk
    )
{
    DWORD ReturnedByteCount;

    return DeviceIoControl(
                hDisk,
                FSCTL_UNLOCK_VOLUME,
                NULL,
                0,
                NULL,
                0,
                &ReturnedByteCount,
                NULL
                );
}

BOOL
DismountVolume(
    HANDLE hDisk
    )
{
    DWORD ReturnedByteCount;

    return DeviceIoControl(
                hDisk,
                FSCTL_DISMOUNT_VOLUME,
                NULL,
                0,
                NULL,
                0,
                &ReturnedByteCount,
                NULL
                );
}


#ifdef JUNK
            BOOL b;
            DWORD BytesRead, BytesWritten;
            DWORD FileSize;
            DWORD VirtBufSize;
            DWORD NumBufs;

            //
            // If we are copying from floppy to file, just do the copy
            // Otherwise, we might have to format the floppy first
            //

            if ( SourceIsDrive ) {

                //
                // Device reads must be sector aligned. VirtualAlloc will
                // garuntee alignment
                //

                NumBufs = Geometry.Cylinders.LowPart;
                VirtBufSize =  Geometry.TracksPerCylinder *
                               Geometry.SectorsPerTrack *
                               Geometry.BytesPerSector;

                IoBuffer = VirtualAlloc(NULL,VirtBufSize,MEM_COMMIT,PAGE_READWRITE);

                if ( !IoBuffer ) {
                    printf("MFMT: Buffer Allocation Failed\n");
                    ExitProcess(1);
                    }

                for ( ;NumBufs > 0; NumBufs-- )
                    {
                    b = ReadFile(hDrive,IoBuffer, VirtBufSize, &BytesRead, NULL);
                    if (b && BytesRead){
                        b = WriteFile(hDiskImage,IoBuffer, BytesRead, &BytesWritten, NULL);
                        if ( !b || ( BytesRead != BytesWritten ) ) {
                            printf("MFMT: Fatal Write Error %d\n",GetLastError());
                            ExitProcess(1);
                            }
                        }
                    else {
                        printf("MFMT: Fatal Read Error %d\n",GetLastError());
                        ExitProcess(1);
                        }
                    }
                }
            else {

                //
                // Check to see if the image will fit on the floppy. If it
                // will, then LowLevelFormat the floppy and press on
                //

                FileSize = GetFileSize(hDiskImage,NULL);

                b = FALSE;
                for(i=0;i<SupportedGeometryCount;i++) {
                    NumBufs = SupportedGeometry[i].Cylinders.LowPart;
                    VirtBufSize =  SupportedGeometry[i].TracksPerCylinder *
                                   SupportedGeometry[i].SectorsPerTrack *
                                   SupportedGeometry[i].BytesPerSector;
                    if ( VirtBufSize*NumBufs >= FileSize ) {

                        IoBuffer = VirtualAlloc(NULL,VirtBufSize,MEM_COMMIT,PAGE_READWRITE);

                        if ( !IoBuffer ) {
                            printf("MFMT: Buffer Allocation Failed\n");
                            ExitProcess(1);
                            }

                        //
                        // Format the floppy
                        //

                        LowLevelFormat(hDrive,&SupportedGeometry[i]);

                        for ( ;NumBufs > 0; NumBufs-- )
                            {
                            b = ReadFile(hDiskImage,IoBuffer, VirtBufSize, &BytesRead, NULL);
                            if (b && BytesRead){
                                b = WriteFile(hDrive,IoBuffer, BytesRead, &BytesWritten, NULL);
                                if ( !b || ( BytesRead != BytesWritten ) ) {
                                    printf("MFMT: Fatal Write Error %d\n",GetLastError());
                                    ExitProcess(1);
                                    }
                                }
                            else {
                                printf("MFMT: Fatal Read Error %d\n",GetLastError());
                                ExitProcess(1);
                                }
                            }
                        b = TRUE;
                        break;
                    	}
                    }

                if ( !b ) {
                    printf("MFMT: FileSize %d is not supported on drive %s\n",FileSize,DriveName);
                    ExitProcess(1);
                    }
                }
        }

        //
        // Dismounting forces the filesystem to re-evaluate the media id
        // and geometry. This is the same as popping the floppy in and out
        // of the disk drive
        //

        DismountVolume(hDrive);
        UnlockVolume(hDrive);

        ExitProcess(0);
        }
    return (0);
}
#endif


#define CHAN_MAX 32
int chan_count = 0;
HANDLE chan_fh[CHAN_MAX];
int chan_secsz[CHAN_MAX];
LPVOID IoBuffer[CHAN_MAX];

unsigned phyio_init(int devlen,char *devnam,unsigned *chan)
{
	unsigned sts = 1;
    if (chan_count < CHAN_MAX - 1 && devlen == 2 &&
        *devnam >= 'A' && *(devnam + 1) == ':') {

		char ntname[20];
        HANDLE hDrive;
        DISK_GEOMETRY Geometry;

        sprintf(ntname,"\\\\.\\%s",devnam);
        chan_fh[chan_count] = hDrive = CreateFile(
                    ntname,
                    GENERIC_READ,
                    0,
                    NULL,
                    OPEN_EXISTING,
                    0,
                    NULL
                    );
        if ( hDrive == INVALID_HANDLE_VALUE ) {
            printf("Open %s failed %d\n",devnam,GetLastError());
            return 64;
            }
        if ( LockVolume(hDrive) == FALSE ) {
            printf("LockVolume %s failed %d\n",devnam,GetLastError());
            return 72;
            }

        if ( !GetDiskGeometry(hDrive,&Geometry) ) {
            printf("GetDiskGeometry %s failed %d\n",devnam,GetLastError());
            return 80;
            }

		chan_secsz[chan_count] = Geometry.BytesPerSector;
        IoBuffer[chan_count] = VirtualAlloc(NULL,chan_secsz[chan_count],MEM_COMMIT,PAGE_READWRITE);
        *chan = chan_count++;

		return 1;

	}

    return 4;
}


unsigned phy_getsect(unsigned chan,unsigned sector,char **buffer)
{
    if (chan < chan_count) {
		BOOL b;
        DWORD BytesRead;
        
		b = SetFilePointer(chan_fh[chan],sector*chan_secsz[chan],0,FILE_BEGIN);

	    b = ReadFile(chan_fh[chan],IoBuffer[chan],chan_secsz[chan],&BytesRead,NULL);
        if (b && BytesRead){
			*buffer = IoBuffer[chan];
            return 1;
		}
	}
    return 64;           
}






unsigned phyio_read(unsigned chan,unsigned block,unsigned length,char *buffer)
{
    register unsigned sts = 1;
    if (chan < chan_count) {
        register unsigned sect = block * 512 / chan_secsz[chan];
        register unsigned offset = (block - sect * chan_secsz[chan] / 512) * 512;
        while (length > 0) {
            register unsigned transfer;
			char *sector;
            if (((sts = phy_getsect(chan,sect,&sector)) & 1) == 0) break;
            transfer = chan_secsz[chan] - offset;
            if (transfer > length) transfer = length;
            memcpy(buffer,sector + offset,transfer);
            buffer += transfer;
            length -= transfer;
            sect++;
            offset = 0;
        }
    } else {
        sts = 64;
    }
    if ((sts & 1) == 0) printf("PHYNT Error %d Block %d Length %d\n",sts,block,length);
    return sts;
}
