/* PHYW95.C  b1.1    Physical I/O module for Windows 95 */

/* Note: This routine uses DeviceIoControl with function
	 VWIN32_DIOC_DOS_INT25.  THIS ONLY WORKS FOR CERTAIN
	 DEVICES!!!!  Floppy drives seem OK but CDs FAIL.
	 If you manage to use BIOS I/O or some other method
	 to get around this problem please send me your code!!! */


#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <Windows.h>
#include <winioctl.h>
#include "phy.h"
#include "ssdef.h"

#define VWIN32_DIOC_DOS_IOCTL 1
#define VWIN32_DIOC_DOS_INT25 2
#define VWIN32_DIOC_DOS_INT26 3

#pragma pack(1)
typedef struct _DIOC_REGISTERS {
    DWORD reg_EBX;
    DWORD reg_EDX;
    DWORD reg_ECX;
    DWORD reg_EAX;
    DWORD reg_EDI;
    DWORD reg_ESI;
    DWORD reg_Flags;
} DIOC_REGISTERS;

typedef struct _DPB {
    unsigned int dpb_sector;
    unsigned short dpb_count;
    char *dpb_buffer;
} DPB;

typedef struct _DEVICEPARAM {
    char junk1[7];
    unsigned short sec_size;
    char junk2[23];
} DEVICEPARAM;

#pragma pack()





#define CHAN_MAX 32
int chan_count = 0;
HANDLE chan_fh[CHAN_MAX];
int chan_drive[CHAN_MAX];
int chan_secsz[CHAN_MAX];


unsigned phyio_init(int devlen,char *devnam,unsigned *chan)
{
    if (chan_count < CHAN_MAX - 1 && devlen == 2 &&
	*devnam >= 'A' && *(devnam + 1) == ':') {
	chan_fh[chan_count] = CreateFile("\\\\.\\vwin32",
					 0,0,
					 (LPSECURITY_ATTRIBUTES) NULL,OPEN_EXISTING,
					 FILE_ATTRIBUTE_NORMAL,(HANDLE) NULL);
#ifdef DEBUG
	printf("Channel: %d\n",chan_fh[chan_count]);
#endif
	if (chan_fh[chan_count] != 0) {
	    DIOC_REGISTERS reg;
	    DEVICEPARAM deviceparam;
	    BOOL fResult;
	    DWORD cb;
	    chan_drive[chan_count] = toupper(*devnam) - 'A';

	    reg.reg_EAX = 0x440d;
	    reg.reg_EBX = 0x000 | chan_drive[chan_count] + 1;
	    reg.reg_ECX = 0x084a;	/* Lock volume */
	    reg.reg_EDX = 0;
	    reg.reg_Flags = 0x0002;	/* Permission  */

	    fResult = DeviceIoControl(chan_fh[chan_count],
				      VWIN32_DIOC_DOS_IOCTL,
				      &reg,sizeof(reg),
				      &reg,sizeof(reg),
				      &cb,0);

	    if (!fResult || (reg.reg_Flags & 0x0001)) {
		printf("Volume lock failed (%d)\n",GetLastError());
		return SS$_DEVNOTALLOC;
	    }
	    reg.reg_EAX = 0x440d;
	    reg.reg_EBX = chan_drive[chan_count] + 1;
	    reg.reg_ECX = 0x0860;	/* Get device parameters */
	    reg.reg_EDX = (DWORD) & deviceparam;
	    reg.reg_Flags = 0x0001;	/* set carry flag  */

	    fResult = DeviceIoControl(chan_fh[chan_count],
				      VWIN32_DIOC_DOS_IOCTL,
				      &reg,sizeof(reg),
				      &reg,sizeof(reg),
				      &cb,0);

	    if (!fResult || (reg.reg_Flags & 0x0001)) {
		printf("Volume get parameters failed (%d)\n",GetLastError());
		deviceparam.sec_size = 2048;
	    }
	    chan_secsz[chan_count] = deviceparam.sec_size;
#ifdef DEBUG
		printf("Sector size is %d\n",deviceparam.sec_size);
#endif

	    *chan = chan_count++;
	    return 1;
	}
    }
    return SS$_IVCHAN;
}


unsigned phy_getsect(unsigned chan,unsigned sector,char *buffer)
{
    if (chan < chan_count) {
	DIOC_REGISTERS reg;
	DPB dpb;
	BOOL fResult;
	DWORD cb;

	dpb.dpb_sector = sector;/* sector number */
	dpb.dpb_count = 1;	/* sector count */
	dpb.dpb_buffer = buffer;/* sector buffer */

	reg.reg_EAX = chan_drive[chan];	/* drive           */
	reg.reg_EBX = (DWORD) & dpb;	/* parameter block */
	reg.reg_ECX = -1;	/* use dpb    */
	reg.reg_EDX = sector;	/* sector num      */
	reg.reg_Flags = 0x0001;	/* set carry flag  */

	fResult = DeviceIoControl(chan_fh[chan],
				  VWIN32_DIOC_DOS_INT25,
				  &reg,sizeof(reg),
				  &reg,sizeof(reg),
				  &cb,0);

	if (!fResult || (reg.reg_Flags & 0x0001)) {
	    printf("Sector %d read failed %d\n",sector,GetLastError());
	    return SS$_PARITY;
	}
	return 1;
    } else {
	return SS$_IVCHAN;
    }
}






unsigned phyio_read(unsigned chan,unsigned block,unsigned length,char *buffer)
{
    register unsigned sts = 1;
    if (chan < chan_count) {
	register unsigned sect = block * 512 / chan_secsz[chan];
	register unsigned offset = (block - sect * chan_secsz[chan] / 512) * 512;
	while (length > 0) {
	    register unsigned transfer;
	    if (offset == 0 && length >= chan_secsz[chan]) {
		transfer = chan_secsz[chan];
		if (((sts = phy_getsect(chan,sect,buffer)) & 1) == 0) break;
	    } else {
		char sector[4096];
		if (((sts = phy_getsect(chan,sect,sector)) & 1) == 0) break;
		transfer = chan_secsz[chan] - offset;
		if (transfer > length) transfer = length;
		memcpy(buffer,sector + offset,transfer);
	    }
	    buffer += transfer;
	    length -= transfer;
	    sect++;
	    offset = 0;
	}
    } else {
	sts = SS$_IVCHAN;
    }
    if ((sts & 1) == 0) printf("PHYW95 Error %d Block %d Length %d\n",sts,block,length);
    return sts;
}
